"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.TypeScriptPlugin = void 0;
const path = require("path");
const fs = require("fs-extra");
const _ = require("lodash");
const globby = require("globby");
const typescript = require("./typescript");
const watchFiles_1 = require("./watchFiles");
const SERVERLESS_FOLDER = '.serverless';
const BUILD_FOLDER = '.build';
// handles broken symlinks
const symlinkExistsSync = (path) => {
    try {
        fs.lstatSync(path);
        return true;
    }
    catch (e) {
        return false;
    }
};
class TypeScriptPlugin {
    constructor(serverless, options) {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l, _m, _o, _p, _q, _r, _s, _t, _u, _v;
        this.serverless = serverless;
        this.options = options;
        this.hooks = {};
        if ((_d = (_c = (_b = (_a = serverless.service) === null || _a === void 0 ? void 0 : _a.custom) === null || _b === void 0 ? void 0 : _b.typescript) === null || _c === void 0 ? void 0 : _c.run) !== null && _d !== void 0 ? _d : true) {
            this.hooks['before:run:run'] = () => __awaiter(this, void 0, void 0, function* () {
                yield this.compileTs();
                yield this.copyExtras();
                yield this.copyDependencies();
            });
        }
        if ((_h = (_g = (_f = (_e = serverless.service) === null || _e === void 0 ? void 0 : _e.custom) === null || _f === void 0 ? void 0 : _f.typescript) === null || _g === void 0 ? void 0 : _g.offline) !== null && _h !== void 0 ? _h : true) {
            this.hooks['before:offline:start'] = () => __awaiter(this, void 0, void 0, function* () {
                yield this.compileTs();
                yield this.copyExtras();
                yield this.copyDependencies();
                this.watchAll();
            });
            this.hooks['before:offline:start:init'] = () => __awaiter(this, void 0, void 0, function* () {
                yield this.compileTs();
                yield this.copyExtras();
                yield this.copyDependencies();
                this.watchAll();
            });
        }
        if ((_m = (_l = (_k = (_j = serverless.service) === null || _j === void 0 ? void 0 : _j.custom) === null || _k === void 0 ? void 0 : _k.typescript) === null || _l === void 0 ? void 0 : _l.package) !== null && _m !== void 0 ? _m : true) {
            this.hooks['before:package:createDeploymentArtifacts'] = () => __awaiter(this, void 0, void 0, function* () {
                yield this.compileTs();
                yield this.copyExtras();
                yield this.copyDependencies(true);
            });
            this.hooks['after:package:createDeploymentArtifacts'] = () => __awaiter(this, void 0, void 0, function* () {
                yield this.cleanup();
            });
        }
        if ((_r = (_q = (_p = (_o = serverless.service) === null || _o === void 0 ? void 0 : _o.custom) === null || _p === void 0 ? void 0 : _p.typescript) === null || _q === void 0 ? void 0 : _q.deploy) !== null && _r !== void 0 ? _r : true) {
            this.hooks['before:deploy:function:packageFunction'] = () => __awaiter(this, void 0, void 0, function* () {
                yield this.compileTs();
                yield this.copyExtras();
                yield this.copyDependencies(true);
            });
            this.hooks['after:deploy:function:packageFunction'] = () => __awaiter(this, void 0, void 0, function* () {
                yield this.cleanup();
            });
        }
        if ((_v = (_u = (_t = (_s = serverless.service) === null || _s === void 0 ? void 0 : _s.custom) === null || _t === void 0 ? void 0 : _t.typescript) === null || _u === void 0 ? void 0 : _u.invoke) !== null && _v !== void 0 ? _v : true) {
            this.hooks['before:invoke:local:invoke'] = () => __awaiter(this, void 0, void 0, function* () {
                var _w, _x, _y, _z;
                if ((_z = (_y = (_x = (_w = serverless.service) === null || _w === void 0 ? void 0 : _w.custom) === null || _x === void 0 ? void 0 : _x.typescript) === null || _y === void 0 ? void 0 : _y.invoke) !== null && _z !== void 0 ? _z : true) {
                    const emitedFiles = yield this.compileTs();
                    yield this.copyExtras();
                    yield this.copyDependencies();
                    if (this.isWatching) {
                        emitedFiles.forEach(filename => {
                            const module = require.resolve(path.resolve(this.originalServicePath, filename));
                            delete require.cache[module];
                        });
                    }
                }
            });
            this.hooks['after:invoke:local:invoke'] = () => {
                if (this.options.watch) {
                    this.watchFunction();
                    this.serverless.cli.log('Waiting for changes...');
                }
            };
        }
    }
    get functions() {
        const { options } = this;
        const { service } = this.serverless;
        if (options.function) {
            return {
                [options.function]: service.functions[this.options.function]
            };
        }
        return service.functions;
    }
    get rootFileNames() {
        return typescript.extractFileNames(this.originalServicePath, this.serverless.service.provider.name, this.functions);
    }
    prepare() {
        // exclude serverless-plugin-typescript
        for (const fnName in this.functions) {
            const fn = this.functions[fnName];
            fn.package = fn.package || {
                exclude: [],
                include: [],
            };
            // Add plugin to excluded packages or an empty array if exclude is undefined
            fn.package.exclude = _.uniq([...fn.package.exclude || [], 'node_modules/serverless-plugin-typescript']);
        }
    }
    watchFunction() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.isWatching) {
                return;
            }
            this.serverless.cli.log(`Watch function ${this.options.function}...`);
            this.isWatching = true;
            watchFiles_1.watchFiles(this.rootFileNames, this.originalServicePath, () => {
                this.serverless.pluginManager.spawn('invoke:local');
            });
        });
    }
    watchAll() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.isWatching) {
                return;
            }
            this.serverless.cli.log(`Watching typescript files...`);
            this.isWatching = true;
            watchFiles_1.watchFiles(this.rootFileNames, this.originalServicePath, this.compileTs.bind(this));
        });
    }
    compileTs() {
        var _a, _b, _c, _d;
        return __awaiter(this, void 0, void 0, function* () {
            this.prepare();
            this.serverless.cli.log('Compiling with Typescript...');
            if (!this.originalServicePath) {
                // Save original service path and functions
                this.originalServicePath = this.serverless.config.servicePath;
                // Fake service path so that serverless will know what to zip
                this.serverless.config.servicePath = path.join(this.originalServicePath, BUILD_FOLDER);
            }
            const tsconfig = typescript.getTypescriptConfig(this.originalServicePath, this.isWatching ? null : this.serverless.cli);
            tsconfig.outDir = BUILD_FOLDER;
            const emitedFiles = yield typescript.run(this.rootFileNames, tsconfig, { paths: (_d = (_c = (_b = (_a = this.serverless.service) === null || _a === void 0 ? void 0 : _a.custom) === null || _b === void 0 ? void 0 : _b.typescript) === null || _c === void 0 ? void 0 : _c.paths) !== null && _d !== void 0 ? _d : false });
            this.serverless.cli.log('Typescript compiled.');
            return emitedFiles;
        });
    }
    /** Link or copy extras such as node_modules or package.include definitions */
    copyExtras() {
        return __awaiter(this, void 0, void 0, function* () {
            const { service } = this.serverless;
            // include any "extras" from the "include" section
            if (service.package.include && service.package.include.length > 0) {
                const files = yield globby(service.package.include);
                for (const filename of files) {
                    const destFileName = path.resolve(path.join(BUILD_FOLDER, filename));
                    const dirname = path.dirname(destFileName);
                    if (!fs.existsSync(dirname)) {
                        fs.mkdirpSync(dirname);
                    }
                    if (!fs.existsSync(destFileName)) {
                        fs.copySync(path.resolve(filename), path.resolve(path.join(BUILD_FOLDER, filename)));
                    }
                }
            }
        });
    }
    /**
     * Copy the `node_modules` folder and `package.json` files to the output
     * directory.
     * @param isPackaging Provided if serverless is packaging the service for deployment
     */
    copyDependencies(isPackaging = false) {
        return __awaiter(this, void 0, void 0, function* () {
            const outPkgPath = path.resolve(path.join(BUILD_FOLDER, 'package.json'));
            const outModulesPath = path.resolve(path.join(BUILD_FOLDER, 'node_modules'));
            // copy development dependencies during packaging
            if (isPackaging) {
                // symlinkExistsSync handles the broken symlink case
                if (fs.existsSync(outModulesPath) || symlinkExistsSync(outModulesPath)) {
                    fs.unlinkSync(outModulesPath);
                }
                if (fs.existsSync(path.resolve('node_modules'))) {
                    fs.copySync(path.resolve('node_modules'), outModulesPath);
                }
            }
            else {
                if (!fs.existsSync(outModulesPath)) {
                    yield this.linkOrCopy(path.resolve('node_modules'), outModulesPath, 'junction');
                }
            }
            // copy/link package.json
            if (!fs.existsSync(outPkgPath)) {
                yield this.linkOrCopy(path.resolve('package.json'), outPkgPath, 'file');
            }
        });
    }
    /**
     * Move built code to the serverless folder, taking into account individual
     * packaging preferences.
     */
    moveArtifacts() {
        return __awaiter(this, void 0, void 0, function* () {
            const { service } = this.serverless;
            yield fs.copy(path.join(this.originalServicePath, BUILD_FOLDER, SERVERLESS_FOLDER), path.join(this.originalServicePath, SERVERLESS_FOLDER));
            if (this.options.function) {
                const fn = service.functions[this.options.function];
                fn.package.artifact = path.join(this.originalServicePath, SERVERLESS_FOLDER, path.basename(fn.package.artifact));
                return;
            }
            if (service.package.individually) {
                const functionNames = service.getAllFunctions();
                functionNames.forEach(name => {
                    service.functions[name].package.artifact = path.join(this.originalServicePath, SERVERLESS_FOLDER, path.basename(service.functions[name].package.artifact));
                });
                return;
            }
            service.package.artifact = path.join(this.originalServicePath, SERVERLESS_FOLDER, path.basename(service.package.artifact));
        });
    }
    cleanup() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.moveArtifacts();
            // Restore service path
            this.serverless.config.servicePath = this.originalServicePath;
            // Remove temp build folder
            fs.removeSync(path.join(this.originalServicePath, BUILD_FOLDER));
        });
    }
    /**
     * Attempt to symlink a given path or directory and copy if it fails with an
     * `EPERM` error.
     */
    linkOrCopy(srcPath, dstPath, type) {
        return __awaiter(this, void 0, void 0, function* () {
            return fs.symlink(srcPath, dstPath, type)
                .catch(error => {
                if (error.code === 'EPERM' && error.errno === -4048) {
                    return fs.copy(srcPath, dstPath);
                }
                throw error;
            });
        });
    }
}
exports.TypeScriptPlugin = TypeScriptPlugin;
module.exports = TypeScriptPlugin;
//# sourceMappingURL=index.js.map